"""
Right, It's uEntity.

For more information see child modules and the `README.md`.
"""
import logging

from untt.entity import Entity  # noqa

logging.getLogger('untt').addHandler(logging.NullHandler())

__version__ = '0.1'



"""
Defines classes for main entity functionality.
"""
from untt import util
from untt.field import Field, EntityField
from untt.util import entity_base, EmptyValue


SCHEMA_URL = 'http://json-schema.org/draft-04/schema#'


class EntityMeta(type):
    """
    The Entity metaclass that defines Entity type creation.

    This scans through the Fields of the Entity class and make appropriate
    setup for further use by the Entity class.
    """
    def __init__(cls, name, bases, dct):
        super(EntityMeta, cls).__init__(name, bases, dct)

        cls.untt_properties = cls._collect_untt_properties(bases, dct)

        title, description = util.parse_docstring(cls.__doc__ or '')
        cls.entity_schema = {
            'type': 'object',
            'properties': {
                n: p.json_schema for n, p in cls.untt_properties.items()
            },
            'required': [n for n, p in cls.untt_properties.items() if not
                         p.optional],
            'title': title,
            'description': description
        }

        if dct.get('schema_root', False):
            cls.schema_definitions = {}

        cls.schema_definitions[cls.__name__] = cls.entity_schema

    @property
    def json_schema(cls):
        """
        Returns the relative reference to the definition ofi its JSON Schema.
        """
        return {
            '$ref': '#/definitions/{}'.format(cls.__name__)
        }

    @classmethod
    def _collect_untt_properties(mcs, bases, dct):
        """
        Prepares a map of Fields defined for the Entity.
        """
        untt_properties = {}

        for base in bases:
            if isinstance(base, EntityMeta):
                untt_properties.update(base.untt_properties)

        untt_properties.update({
            name: p for name, p in dct.items()
            if isinstance(p, Field)
        })

        return untt_properties


@entity_base
class Entity(object, metaclass=EntityMeta):
    """
    Entity is the main class of untt.

    To define an entity users should subclass Entity and define its fileds
    using the provided types in `untt.types`, e.g.:

        class MyEntity(Entity):
            my_int_field = Integer()
            my-str_field = String()

    The JSON schema of the entity will be accessable as
    `MyEntity.entity_schema`. All the JSON schemas of the entities subclassing
    the `Entity` will be accessible via `Entity.schema_definitions`.

    When defining a base class for other entities, you may decorate it with the
    `untt.util.entity_base` to exclude it from all definitions. Also, to make
    a base class the root of shcema definitions and start from scratch, assign
    the class variable `schema_root` of the base class the value `True`.

    In order to define any field of the entity as another entity type, you may
    use the `as_property` class method, e.g.:

        class MyOtherEntity(Entity):
            my_entity_field = MyEntity.as_property()

    Finally, you will get a bonus functions `to_json` and `from_json` to
    convert your entity objects to/from JSON.
    """
    schema_root = True
    untt_properties = {}
    schema_definitions = {}
    entity_schema = {}
    json_schema = {}
    not_provided = EmptyValue()

    def to_json(self):
        """
        Convert `Entity` object to JSON object.
        """
        json_value = {}
        for name, prop in self.untt_properties.items():
            value = getattr(self, name)
            json_value[name] = prop.to_json(value)

        return json_value

    @classmethod
    def as_property(cls, *,
                    optional=False,
                    nullable=False):
        """
        Returns an `EntityField` instance representing the owning entity type.
        """
        return EntityField(cls,
                           optional=optional,
                           nullable=nullable)

    @classmethod
    def load_json(cls, json_obj):
        """
        Load a json object key-by-key.
        """
        cls.validate(json_obj)

        parsed_json_obj = {}
        for name, prop in cls.untt_properties.items():
            value = json_obj.get(name, cls.not_provided)

            if value is cls.not_provided and prop.default is not prop.empty:
                value = prop.default

            parsed_json_obj[name] = prop.from_json(value)

        return parsed_json_obj

    @classmethod
    def from_json(cls, json_value):
        """
        Initiate a `cls` instance out of a JSON object.
        """
        json_obj = cls.load_json(json_value)

        try:
            obj = cls()
        except TypeError as ex:
            raise TypeError(
                "'{}' implements a custom '__init__' with arguments. "
                "Please implement a custom 'from_json' to support it."
                .format(cls.__name__)
            ) from ex

        for name, value in json_obj.items():
            setattr(obj, name, value)

        return obj

    @classmethod
    def validate(cls, json_value):
        """
        Validate a given `json_value` against the `cls` JSON schema.
        """
        schema = {
            '$schema': SCHEMA_URL,
            'definitions': cls.schema_definitions,
        }
        schema.update(cls.json_schema)
        util.validate(json_value, schema)



"""
This module defines the root classes to implement entity fields.
"""
from untt.ex import ValidationError


class Field(object):
    """
    This is the root class for defining Entity fields.

    This implements the Descriptor protocol, and manages the values of the
    properties by validating the values being assigned to them. A subclasse
    should implement the `_validate` method for that.

    This class also provides the ability to convert other non-primitive types
    from and to JSON that are originally not serializable. To enable this onc
    may first specify the `_python_type` they expect. If specified, an
    additional instance type check will be performed. Next, the methods
    `from_json` and `to_json` methods should be extended.
    """
    _python_type = None

    def __init__(self, *, title="",
                 description="",
                 optional=False,
                 nullable=False):
        self.title = title
        self.description = description
        self.optional = optional
        self.nullable = nullable

        self._values = {}

    def __get__(self, owner, klass=None):
        if owner is None:
            return self
        else:
            return self._values[owner]

    def __set__(self, obj, value):
        if self._python_type:
            if not isinstance(value, self._python_type):
                raise ValidationError(
                    "'{}' is not a '{}' object".format(
                        value, self._python_type
                    )
                )

        json_value = self.to_json(value)
        self._validate(json_value)

        self._values[obj] = value

    def _validate(self, json_value):
        """
        Should raise ValidationError if `json_value` does not fit the shcema.
        """
        raise NotImplementedError()  # pragma: no cover

    def from_json(self, json_value):
        """
        Convert `json_value` to appropriate type.
        """
        return json_value

    def to_json(self, value):
        """
        Convert the `value` to JSON.
        """
        return value


class EntityField(Field):
    """
    A Field type that represents an `Entity` implementation.
    """
    def __init__(self, entity_type, *,
                 title="",
                 description="",
                 optional=False,
                 nullable=False):
        self.entity_type = entity_type
        self._python_type = entity_type

        super(EntityField, self).__init__(title=title,
                                          description=description,
                                          optional=optional,
                                          nullable=nullable)

    @property
    def json_schema(self):
        """
        The JSON schema definition for the field, i.e. entity.
        """
        return self.entity_type.entity_schema

    def from_json(self, json_value):
        return self.entity_type.from_json(json_value)

    def to_json(self, value):
        return value.to_json()

    def _validate(self, json_value):
        self.entity_type.validate(json_value)



"""
Provides various utility functions and decorators.
"""
from functools import wraps

from jsonschema import (validate as jsonschema_validate,
                        ValidationError as JsonSchemaValidationError,
                        FormatChecker)

from untt.ex import ValidationError


def normalize_docsplit(docsplit):
    """
    Strips the lines and the line list.
    """
    docsplit = [l.strip() for l in docsplit]

    if len(docsplit) > 0:
        if docsplit[0] == '':
            docsplit = docsplit[1:]

    if len(docsplit) > 0:
        if docsplit[-1] == '':
            docsplit = docsplit[:-1]

    return docsplit


def parse_docstring(doc):
    """
    Parses title and description out of a docstring.
    """
    doc_split = normalize_docsplit(doc.split('\n'))
    if len(doc_split) == 0:
        return '', ''
    elif len(doc_split) == 1:
        return doc_split[0], ''

    if doc_split[1] != '':
        return '', '\n'.join(doc_split)
    else:
        title = doc_split[0]
        description = '\n'.join(doc_split[2:])

        return title, description


def entity_base(klass):
    """
    Indicates that an `Entity` is a base and not an actual one.

    Entities decorated with this decorator will not be included in the root
    schema definitions.
    """
    del klass.schema_definitions[klass.__name__]
    return klass


@wraps(jsonschema_validate)
def validate(*args, **kwargs):
    """
    jsonschema.validate wrapper, that raises `untt.ex.ValidationError`.
    """
    try:
        jsonschema_validate(*args, **kwargs, format_checker=FormatChecker())
    except JsonSchemaValidationError as ex:
        raise ValidationError(str(ex)) from ex


class EmptyValue(object):
    """Used to represent an empty dummy value."""
    pass



"""
Defines primitive field types
"""
import datetime

import iso8601

from untt.field import Field
from untt import util
from untt.ex import DefinitionError

from untt.util import EmptyValue


class PrimitiveType(Field):
    """
    A Field type that represents a primitive type, defined by JSON Schema.
    """
    _ACCEPTABLE_TYPES = ('string', 'boolean', 'integer',
                         'number', 'null', 'array')
    empty = EmptyValue()

    def __init__(self, property_type, *,
                 title="", description="",
                 default=empty, nullable=False,
                 **keywords):
        if property_type not in self._ACCEPTABLE_TYPES:
            raise DefinitionError(
                "A property must be of one of the following types: %s",
                self._ACCEPTABLE_TYPES
            )

        self.property_type = property_type
        self.default = default
        self.keywords = keywords

        super(PrimitiveType, self).__init__(title=title,
                                            description=description,
                                            optional=default is not self.empty,
                                            nullable=nullable)

    @property
    def json_schema(self):
        """
        The JSON Schema definition for the primitive type.
        """
        schema = {
            'type': self.property_type,

            'title': self.title,
            'description': self.description,

            **self.keywords
        }

        if self.default is not self.empty:
            schema['default'] = self.default

        if self.nullable:
            schema = {
                'oneOf': [
                    schema,
                    {'type': 'null'}
                ]
            }

        return schema

    def _validate(self, json_value):
        util.validate(json_value, self.json_schema)


class BaseType(PrimitiveType):
    """
    Base class for easy primitive type definition.
    """
    property_type = None
    keywords = {}

    def __init__(self, *, title="", description="",
                 default=PrimitiveType.empty, nullable=False,
                 **keywords):
        self.keywords.update(keywords)
        super(BaseType, self).__init__(self.property_type, title=title,
                                       description=description,
                                       nullable=nullable, default=default,
                                       **self.keywords)


class Integer(BaseType):
    """
    JSON Schema integer type.
    """
    property_type = 'integer'


class Number(BaseType):
    """
    JSON Schema number type.
    """
    property_type = 'number'


class String(BaseType):
    """
    JSON Schema string type.
    """
    property_type = 'string'


class Array(BaseType):
    """
    JSON Schema array type.
    """
    property_type = 'array'
    keywords = {
        'items': {}
    }


class Boolean(BaseType):
    """
    JSON Schema boolean type.
    """
    property_type = 'boolean'


class Datetime(String):
    """
    JSON Schema string type with date-time format.

    This accepts a `datetime` Python object as a value and converts back and
    forth automatically.
    """
    _python_type = datetime.datetime
    keywords = {
        'format': 'date-time'
    }

    def to_json(self, value):
        return value.isoformat()

    def from_json(self, json_value):
        return iso8601.parse_date(json_value)


class Email(String):
    """
    JSON Schema string type with email format.
    """
    keywords = {
        'format': 'email'
    }



from os import path
import codecs

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

with codecs.open(path.join(here, 'requirements.txt'),
                 encoding='utf-8') as reqs:
    requirements = reqs.read()

setup(
    name='untt',
    version='0.1',

    description='Right, it\'s uEntity.',
    long_description="""
    """,

    url='http://untt.n9co.de',

    author='Bagrat Aznauryan',
    author_email='bagrat@aznauryan.org',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',

        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],

    keywords='model orm entity untt',

    packages=find_packages(exclude=['docs', 'tests']),

    install_requires=requirements,
)



"""
This module is where the untt error definitions live.
"""


class UnttError(Exception):
    """
    Root error class for untt.
    """
    pass


class DefinitionError(UnttError):
    """
    Raises when there is an entity definition error.
    """
    pass


class ValidationError(UnttError):
    """
    Raises when there is a validation error.
    """
    pass
